"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  default: () => web_default
});
module.exports = __toCommonJS(src_exports);

// src/helper/simpleHash.ts
function simpleHash(input = String(Math.random()), length = 3) {
  let hash = 0;
  if (input.length === 0)
    return "";
  for (let i = 0; i < input.length; i++) {
    const char = input.charCodeAt(i);
    hash = (hash << 5) - hash + char;
    hash &= hash;
  }
  return Math.abs(hash).toString(36).substring(0, length);
}

// src/context.ts
var CONTEXT = /* @__PURE__ */ new WeakMap();
var Context = class _Context {
  constructor(path, parent, name) {
    this.define = {};
    this.macros = {};
    this.uid = "";
    this.also = /* @__PURE__ */ new Set();
    this.props = /* @__PURE__ */ new Map();
    this.usedBy = /* @__PURE__ */ new Set();
    this.children = /* @__PURE__ */ new Set();
    CONTEXT.set(path, this);
    this.path = path;
    if (!(parent instanceof _Context))
      return;
    this.parent = parent;
    this.uid = name + "_" + simpleHash(parent.uid);
    this.define = Object.create(parent.define);
    this.macros = Object.create(parent.macros);
    do
      if (parent.condition) {
        parent.children.add(this);
      }
    while (parent = parent.parent);
  }
  static get(from) {
    return CONTEXT.get(from);
  }
};

// src/helper/errors.ts
var Status = {
  currentFile: void 0
};
function ParseErrors(register) {
  const Errors = {};
  for (const error in register) {
    const message = [];
    for (const segment of register[error].split(/\{(?=\d+\})/)) {
      const ammend = /(\d+)\}(.*)/.exec(segment);
      if (ammend)
        message.push(parseInt(ammend[1]), ammend[2]);
      else
        message.push(segment);
    }
    Errors[error] = (node, ...args) => {
      let quote = "";
      if ("node" in node)
        node = node.node;
      for (const slice of message)
        quote += typeof slice == "string" ? slice : args[slice - 1];
      return Status.currentFile.buildCodeFrameError(node, quote);
    };
  }
  return Errors;
}
function parseError(path, err, modiferName) {
  if (!(err instanceof Error))
    return path.hub.buildError(path.node, `Modifier "${modiferName}" failed: ${err}`);
  const stack = err.stack.split("\n    at ");
  const message = err instanceof Error ? err.message : err;
  const error = path.hub.buildError(path.node, `Modifier "${modiferName}" failed: ${message}`);
  error.stack = stack.slice(0, stack.findIndex((line) => /^parse/.test(line)) + 1).join("\n    at ");
  return error;
}

// src/types/assert.ts
var assert_exports = {};
__export(assert_exports, {
  and: () => and,
  anti: () => anti,
  falsy: () => falsy,
  inverseExpression: () => inverseExpression,
  isBinaryAssertion: () => isBinaryAssertion,
  isFalsy: () => isFalsy,
  isParenthesized: () => isParenthesized,
  ternary: () => ternary,
  truthy: () => truthy
});
var ASSERT_OP = /* @__PURE__ */ new Set([
  "in",
  "instanceof",
  "==",
  "===",
  "!=",
  "!==",
  ">",
  ">=",
  "<",
  "<="
]);
var INVERSE_OP = /* @__PURE__ */ new Map([
  ["==", "!="],
  ["===", "!=="],
  [">", "<="],
  ["<", ">="]
]);
for (const [a, b] of INVERSE_OP)
  INVERSE_OP.set(b, a);
function isParenthesized(node) {
  const { extra } = node;
  return extra ? extra.parenthesized === true : false;
}
function isBinaryAssertion(exp) {
  if (types_default.isBinaryExpression(exp) && ASSERT_OP.has(exp.operator))
    return true;
  return false;
}
function inverseExpression(exp) {
  const inverse = INVERSE_OP.get(exp.operator);
  if (inverse)
    return types_default.binaryExpression(inverse, exp.left, exp.right);
  throw new Error(`Can't invert binary comparison ${exp.operator}.`);
}
function isFalsy(exp) {
  return types_default.isUnaryExpression(exp, { operator: "!" });
}
function falsy(exp) {
  return isBinaryAssertion(exp) ? inverseExpression(exp) : types_default.unaryExpression("!", exp, true);
}
function and(a, b) {
  return types_default.logicalExpression("&&", a, b);
}
function anti(exp) {
  return isFalsy(exp) ? exp.argument : falsy(exp);
}
function truthy(a) {
  return isFalsy(a) || isBinaryAssertion(a) ? a : falsy(falsy(a));
}
function ternary(test, consequent, alternate) {
  return types_default.conditionalExpression(test, consequent, alternate);
}

// src/types/construct.ts
var construct_exports = {};
__export(construct_exports, {
  block: () => block,
  call: () => call,
  declare: () => declare,
  get: () => get,
  keyIdentifier: () => keyIdentifier,
  literal: () => literal,
  member: () => member,
  object: () => object,
  objectAssign: () => objectAssign,
  objectKeys: () => objectKeys,
  property: () => property,
  requires: () => requires,
  returns: () => returns,
  statement: () => statement,
  templateExpression: () => templateExpression
});
function literal(value) {
  switch (typeof value) {
    case "string":
      return types_default.stringLiteral(value);
    case "number":
      return types_default.numericLiteral(value);
    case "boolean":
      return types_default.booleanLiteral(value);
    case "undefined":
      return types_default.identifier("undefined");
    case "object":
      if (value === null)
        return types_default.nullLiteral();
    default:
      throw new Error("Not a literal type");
  }
}
function keyIdentifier(name) {
  return /^[A-Za-z0-9$_]+$/.test(name) ? types_default.identifier(name) : types_default.stringLiteral(name);
}
function property(key, value) {
  let shorthand = false;
  if (typeof key == "string") {
    shorthand = types_default.isIdentifier(value, { name: key });
    key = keyIdentifier(key);
  }
  return types_default.objectProperty(key, value, false, shorthand);
}
function object(obj = {}) {
  let properties = [];
  if (Array.isArray(obj))
    properties = obj;
  else
    for (const [key, value] of Object.entries(obj))
      if (value)
        properties.push(property(key, value));
  return types_default.objectExpression(properties);
}
function get(object2, ...path) {
  if (object2 == "this")
    object2 = types_default.thisExpression();
  if (typeof object2 == "string")
    path = [...object2.split("."), ...path];
  for (const x of path) {
    let select;
    if (typeof x == "number")
      select = literal(x);
    else if (typeof x == "string")
      select = keyIdentifier(x);
    else if (types_default.isExpression(x))
      select = x;
    else
      throw new Error("Bad member id, only strings and numbers are allowed");
    object2 = typeof object2 == "object" ? member(object2, select) : select;
  }
  return object2;
}
function member(object2, property2) {
  return types_default.memberExpression(object2, property2, !types_default.isIdentifier(property2));
}
function call(callee, ...args) {
  if (typeof callee == "string")
    callee = get(callee);
  return types_default.callExpression(callee, args);
}
function requires(from) {
  return call("require", literal(from));
}
function returns(argument, parenthesized = false) {
  const statement2 = types_default.returnStatement(argument);
  statement2.extra = { parenthesized };
  return statement2;
}
function declare(kind, id, init) {
  return types_default.variableDeclaration(kind, [
    types_default.variableDeclarator(id, init || null)
  ]);
}
function objectAssign(...objects) {
  return call("Object.assign", ...objects);
}
function objectKeys(object2) {
  return call("Object.keys", object2);
}
function templateExpression(text) {
  return types_default.templateLiteral([
    types_default.templateElement({ raw: text, cooked: text }, false)
  ], []);
}
function statement(from) {
  return types_default.isExpression(from) ? types_default.expressionStatement(from) : from;
}
function block(...statements) {
  const stats = statements.map(statement);
  return types_default.blockStatement(stats);
}

// src/types/jsx.ts
var jsx_exports = {};
__export(jsx_exports, {
  jsxTag: () => jsxTag
});
function jsxIdentifier(name) {
  return typeof name == "string" ? types_default.jsxIdentifier(name) : name;
}
function jsxTag(tag, props, children) {
  const type = jsxIdentifier(tag);
  const content = children.map(jsxContent);
  const contains = content.length > 0;
  const openingElement = types_default.jsxOpeningElement(type, props, !contains);
  const closingElement = contains ? types_default.jsxClosingElement(type) : null;
  return types_default.jsxElement(openingElement, closingElement, content, !contains);
}
function jsxContent(child) {
  if (types_default.isJSXElement(child))
    return child;
  if (types_default.isStringLiteral(child) && !/\{/.test(child.value))
    return types_default.jsxText(child.value);
  return types_default.jsxExpressionContainer(child);
}

// src/types/index.ts
var babel = {};
var types_default = {
  ...babel,
  ...assert_exports,
  ...construct_exports,
  ...jsx_exports
};

// src/syntax/arguments.ts
var Oops = ParseErrors({
  UnaryUseless: "Unary operator here doesn't do anything",
  HexNoNegative: `Hexadecimal numbers are converted into colors (#FFF) so negative sign doesn't mean anything here.
Parenthesize the number if you really need "-{1}" for some reason...`,
  ArrowNotImplemented: "Arrow Function not supported yet",
  ArgumentSpread: "Cannot parse argument spreads for modifier handlers",
  UnknownArgument: "Unknown argument while parsing for modifier.",
  MustBeIdentifier: "Only Identifers allowed here! Call expression must reference another modifier.",
  ModiferCantParse: "Illegal value in modifier",
  ElseNotSupported: "An else statement in an if modifier is not yet supported"
});
function parseArgument(element, childKey) {
  const exp = element.get("expression").node;
  const args = parse.Expression(exp, childKey);
  return Array.isArray(args) ? args : [args];
}
var parse = {
  Expression(element, childKey) {
    if (childKey)
      element = element[childKey];
    if (types_default.isParenthesized(element))
      return element;
    if (element.type in this)
      return this[element.type](element);
    throw Oops.UnknownArgument(element);
  },
  Identifier(node) {
    return node.name;
  },
  BooleanLiteral(bool) {
    return bool.value;
  },
  StringLiteral(e) {
    if (e.value === "")
      return '""';
    return e.value;
  },
  NullLiteral() {
    return null;
  },
  TemplateLiteral(temp) {
    if (temp.quasis.length == 1)
      return temp.quasis[0].value.raw;
    return temp;
  },
  NumericLiteral(number, negative) {
    let { extra: { rawValue, raw } } = number;
    if (types_default.isParenthesized(number) || !/^0x/.test(raw)) {
      if (raw.indexOf(".") > 0)
        return negative ? "-" + raw : raw;
      return negative ? -rawValue : rawValue;
    }
    if (negative)
      throw Oops.HexNoNegative(number, rawValue);
    return HEXColor(raw);
  },
  UnaryExpression(e) {
    const { argument, operator } = e;
    if (operator == "-" && types_default.isNumericLiteral(argument))
      return this.NumericLiteral(argument, true);
    if (operator == "!" && types_default.isIdentifier(argument, { name: "important" }))
      return "!important";
    throw Oops.UnaryUseless(e);
  },
  BinaryExpression(binary) {
    const { left: left2, right: right2, operator } = binary;
    if (operator == "-" && types_default.isIdentifier(left2) && types_default.isIdentifier(right2, { start: left2.end + 1 }))
      return left2.name + "-" + right2.name;
    return [
      operator,
      this.Expression(binary, "left"),
      this.Expression(binary, "right")
    ];
  },
  SequenceExpression(sequence) {
    return sequence.expressions.map((x) => this.Expression(x));
  },
  CallExpression(e) {
    const callee = e.callee;
    const args = [];
    if (callee.type !== "Identifier")
      throw Oops.MustBeIdentifier(callee);
    for (const item of e.arguments) {
      if (types_default.isSpreadElement(item))
        throw Oops.ArgumentSpread(item);
      if (!types_default.isExpression(item))
        throw Oops.UnknownArgument(item);
      args.push(this.Expression(item));
    }
    const { name } = callee;
    if (CSS_UNITS.has(name))
      return args.map((x) => String(x) + name).join(" ");
    return callee.name + `(${args.join(", ")})`;
  },
  ArrowFunctionExpression(e) {
    throw Oops.ArrowNotImplemented(e);
  }
};
var CSS_UNITS = /* @__PURE__ */ new Set([
  "ch",
  "cm",
  "em",
  "ex",
  "in",
  "mm",
  "pc",
  "pt",
  "px",
  "rem",
  "vh",
  "vmax",
  "vmin",
  "vw"
]);
function HEXColor(raw) {
  raw = raw.substring(2);
  if (raw.length == 1)
    raw = "000" + raw;
  else if (raw.length == 2)
    raw = "000000" + raw;
  if (raw.length % 4 == 0) {
    let decimal = [];
    if (raw.length == 4)
      decimal = Array.from(raw).map((x) => parseInt(x + x, 16));
    else
      for (let i = 0; i < 4; i++)
        decimal.push(
          parseInt(raw.slice(i * 2, i * 2 + 2), 16)
        );
    decimal[3] = (decimal[3] / 255).toFixed(3);
    return `rgba(${decimal.join(", ")})`;
  }
  return "#" + raw;
}

// src/syntax/names.ts
function getName(path) {
  let encounteredReturn;
  while (path)
    switch (path.type) {
      case "LabeledStatement":
        return path.node.label.name;
      case "VariableDeclarator": {
        const { id } = path.node;
        return types_default.isIdentifier(id) ? id.name : path.parentPath.node.kind;
      }
      case "AssignmentExpression":
      case "AssignmentPattern": {
        const { left: left2 } = path.node;
        return types_default.isIdentifier(left2) ? left2.name : "assignment";
      }
      case "FunctionDeclaration":
        return path.node.id.name;
      case "ExportDefaultDeclaration":
        return getLocalFilename(path.hub);
      case "ArrowFunctionExpression": {
        path = path.parentPath;
        continue;
      }
      case "ReturnStatement": {
        if (encounteredReturn)
          return "return";
        encounteredReturn = path;
        const ancestry = path.getAncestry();
        const within = ancestry.find((x) => x.isFunction());
        const { node } = within;
        if ("id" in node && node.id)
          return node.id.name;
        if (types_default.isObjectMethod(node)) {
          path = within.getAncestry()[2];
          continue;
        }
        if (types_default.isClassMethod(node)) {
          if (node.key.type !== "Identifier")
            return "ClassMethod";
          if (node.key.name != "render")
            return node.key.name;
          const owner = within.parentPath.parentPath;
          if (owner.node.id)
            return owner.node.id.name;
          path = owner.parentPath;
          continue;
        }
        path = within.parentPath;
        continue;
      }
      case "ObjectProperty": {
        const { key } = path.node;
        return types_default.isIdentifier(key) ? key.name : types_default.isStringLiteral(key) ? key.value : "property";
      }
      default:
        return "element";
    }
  return "element";
}
function getLocalFilename(hub) {
  try {
    const { basename, dirname, sep: separator } = require("path");
    const url = hub.file.opts.filename;
    const [base] = basename(url).split(".");
    if (base !== "index")
      return base;
    return dirname(url).split(separator).pop();
  } catch (err) {
    return "File";
  }
}
function uniqueIdentifier(scope, name = "temp") {
  let uid = name;
  let i = 0;
  while (scope.hasLabel(uid) || scope.hasBinding(uid) || scope.hasGlobal(uid)) {
    uid = name + ++i;
  }
  ;
  if (i > 1)
    uid = name + i;
  return types_default.identifier(uid);
}

// src/label.ts
function handleLabel(path) {
  const context = getContext(path);
  const body = path.get("body");
  let { name } = path.node.label;
  if (body.isBlockStatement()) {
    context.define[name] = new Context(path, context, name);
    return;
  }
  if (!body.isExpressionStatement())
    throw parseError(body, "Not an expression", name);
  if (!context)
    throw parseError(body, "Missing context", name);
  const args = parseArgument(body);
  try {
    const queue = [{ name, args }];
    while (queue.length) {
      const { name: name2, args: args2 } = queue.pop();
      const macro = context.macros[name2];
      const apply = (args3) => {
        context.props.set(name2, args3);
      };
      if (!macro) {
        apply(args2);
        continue;
      }
      const output = macro.apply(context, args2);
      if (!output)
        continue;
      if (Array.isArray(output)) {
        apply(output);
        continue;
      }
      if (typeof output != "object")
        throw new Error("Invalid modifier output.");
      for (const key in output) {
        let args3 = output[key];
        if (args3 === void 0)
          continue;
        if (!Array.isArray(args3))
          args3 = [args3];
        if (key === name2)
          apply(args3);
        else
          queue.push({ name: key, args: args3 });
      }
    }
  } catch (err) {
    throw parseError(body, err, name);
  }
}
function getContext(path) {
  let key = path.key;
  while (path = path.parentPath) {
    const context = Context.get(path);
    if (context instanceof Context) {
      if (key === "alternate") {
        let { alternate, parent, path: path2 } = context;
        if (!alternate) {
          alternate = new Context(path2, parent, "else");
          context.children.add(alternate);
          context.alternate = alternate;
        }
        return alternate;
      }
      return context;
    }
    if (path.isFunction())
      return createFunctionContext(path);
    if (path.isIfStatement())
      return createIfContext(path);
    key = path.key;
  }
  throw new Error("Context not found");
}
function createFunctionContext(path) {
  const name = getName(path);
  const context = getContext(path);
  const component = new Context(path, context, name);
  const body = path.get("body");
  component.define["this"] = component;
  onExit(path, () => {
    if (body.isBlockStatement() && !body.get("body").length)
      body.pushContainer("body", types_default.expressionStatement(
        types_default.jsxElement(
          types_default.jsxOpeningElement(
            types_default.jsxIdentifier("this"),
            [],
            true
          ),
          null,
          [],
          true
        )
      ));
  });
  return component;
}
function createIfContext(path) {
  const test = path.node.test;
  const name = types_default.isIdentifier(test) ? test.name : "if";
  const outer = getContext(path);
  const inner = new Context(path, outer, name);
  if (types_default.isStringLiteral(test)) {
    outer.children.add(inner);
    inner.uid = outer.uid;
    inner.condition = test.value;
  } else {
    outer.also.add(inner);
    inner.condition = test;
  }
  onExit(path, (path2, key) => {
    if (key == "alternate" || inner.alternate)
      return;
    if (!path2.removed)
      path2.remove();
  });
  return inner;
}

// src/syntax/jsx.ts
function getNames(path) {
  const names = /* @__PURE__ */ new Map();
  const opening = path.get("openingElement");
  let tag = opening.get("name");
  while (tag.isJSXMemberExpression()) {
    names.set(tag.get("property").toString(), tag);
    tag = tag.get("object");
  }
  if (tag.isJSXIdentifier())
    names.set(tag.toString(), tag);
  opening.get("attributes").forEach((attr) => {
    if (!attr.isJSXAttribute() || attr.node.value)
      return;
    let { name } = attr.node.name;
    if (typeof name !== "string")
      name = name.name;
    names.set(name, attr);
  });
  return names;
}
function fixImplicitReturn(path) {
  let { node, parentPath: parent } = path;
  if (!parent.isExpressionStatement())
    return false;
  const block2 = parent.parentPath;
  if (block2.isBlockStatement() && block2.get("body").length == 1 && block2.parentPath.isArrowFunctionExpression())
    block2.replaceWith(types_default.parenthesizedExpression(node));
  else
    parent.replaceWith(types_default.returnStatement(node));
  path.skip();
  return true;
}
function spreadProps(path, props) {
  path.get("openingElement").unshiftContainer("attributes", types_default.jsxSpreadAttribute(props));
}
function setTagName(path, name) {
  const { openingElement, closingElement } = path.node;
  const tag = types_default.jsxIdentifier(name);
  openingElement.name = tag;
  if (closingElement)
    closingElement.name = tag;
}
function hasProp(path, name) {
  for (const attr of path.node.openingElement.attributes)
    if (types_default.isJSXAttribute(attr) && attr.name.name === name) {
      const { value } = attr;
      if (types_default.isJSXExpressionContainer(value) && types_default.isExpression(value.expression))
        return value.expression;
      if (types_default.isExpression(value))
        return value;
    }
}

// src/plugin.ts
var HANDLED = /* @__PURE__ */ new WeakMap();
function Plugin(_compiler, options) {
  const SCOPE = /* @__PURE__ */ new WeakMap();
  const { apply } = options;
  if (!apply)
    throw new Error(`Plugin has not defined an apply method.`);
  return {
    manipulateOptions(_options, parse2) {
      parse2.plugins.push("jsx");
    },
    visitor: {
      Program(path, state) {
        const context = new Context(path);
        context.uid = simpleHash(state.filename);
        context.define = Object.assign({}, ...options.define || []);
        context.macros = Object.assign({}, ...options.macros || []);
      },
      BlockStatement: {
        exit
      },
      LabeledStatement: {
        enter(path) {
          const body = path.get("body");
          if (body.isFor() || body.isWhile())
            return;
          handleLabel(path);
          onExit(path, () => {
            if (!path.removed)
              path.remove();
          });
        },
        exit
      },
      JSXElement(path, state) {
        if (path.getData("handled"))
          return;
        if (fixImplicitReturn(path))
          return;
        const parent = path.parentPath;
        const context = !parent.isJSXElement() && getContext(parent);
        const scope = new Set(context ? [context] : SCOPE.get(parent));
        const using = /* @__PURE__ */ new Set();
        SCOPE.set(path, scope);
        getNames(path).forEach((attr, name) => {
          let used = false;
          for (let { define } of scope) {
            const apply2 = [];
            for (let mod; mod = define[name]; define = Object.getPrototypeOf(define)) {
              apply2.push(mod, ...mod.also);
              used = true;
              if (name == "this")
                break;
            }
            apply2.reverse().forEach((ctx) => {
              ctx.usedBy.add(path);
              scope.add(ctx);
              using.add(ctx);
            });
          }
          if (used && attr.isJSXAttribute())
            attr.remove();
        });
        apply(path, using, state);
        if (context === false || context.define.this !== context || context.props.size === 0 || context.usedBy.size)
          return;
        const [inserted] = path.replaceWith(
          types_default.jsxElement(
            types_default.jsxOpeningElement(types_default.jSXIdentifier("this"), []),
            types_default.jsxClosingElement(types_default.jSXIdentifier("this")),
            [path.node]
          )
        );
        apply(inserted, [context], state);
        inserted.setData("handled", true);
      }
    }
  };
}
var plugin_default = Plugin;
function onExit(path, callback) {
  HANDLED.set(path, callback);
}
function exit(path) {
  for (const p of path.getAncestry()) {
    if (p.isBlockStatement())
      continue;
    const callback = HANDLED.get(p);
    if (callback)
      callback(p, p.key);
    else
      break;
  }
}

// src/syntax/component.ts
function componentProp(path, name) {
  const func = path.find((x) => x.isFunction());
  let [props] = func.node.params;
  if (types_default.isObjectPattern(props)) {
    const { properties } = props;
    const prop = properties.find((x) => types_default.isObjectProperty(x) && types_default.isIdentifier(x.key, { name }));
    if (prop)
      return prop.value;
    const id = types_default.identifier(name);
    properties.unshift(
      types_default.objectProperty(id, id, false, true)
    );
    return id;
  } else if (!props) {
    props = uniqueIdentifier(path.scope, "props");
    func.unshiftContainer("params", props);
  }
  if (types_default.isIdentifier(props))
    return types_default.memberExpression(props, types_default.identifier(name));
  throw new Error(`Expected an Identifier or ObjectPattern, got ${props.type}`);
}
function componentProps(path) {
  const func = path.find((x) => x.isFunction());
  let [props] = func.node.params;
  let output;
  if (!props) {
    func.pushContainer("params", output = uniqueIdentifier(path.scope, "props"));
  } else if (types_default.isObjectPattern(props)) {
    const existing = props.properties.find((x) => types_default.isRestElement(x));
    if (types_default.isRestElement(existing))
      output = existing.argument;
    const inserted = types_default.restElement(uniqueIdentifier(path.scope, "rest"));
    props.properties.push(inserted);
    output = inserted.argument;
  } else
    output = props;
  if (types_default.isIdentifier(output))
    return output;
  throw new Error("Could not extract props from function.");
}

// src/syntax/program.ts
var POLYFILL_DEFAULT = "@expressive/babel-preset/polyfill";
function getHelper(name, from, options) {
  const { polyfill = POLYFILL_DEFAULT } = options;
  const program = from.findParent((path) => path.isProgram());
  const body = program.get("body");
  for (const statement2 of body) {
    if (!statement2.isImportDeclaration() || statement2.node.source.value !== polyfill)
      continue;
    const specifiers = statement2.get("specifiers");
    for (const spec of specifiers) {
      if (!spec.isImportSpecifier())
        continue;
      if (types_default.isIdentifier(spec.node.imported, { name }))
        return spec.node.local;
    }
  }
  const concat = uniqueIdentifier(from.scope, name);
  if (polyfill) {
    const importStatement = types_default.importDeclaration(
      [types_default.importSpecifier(concat, types_default.identifier(name))],
      types_default.stringLiteral(polyfill)
    );
    program.unshiftContainer("body", importStatement);
  }
  return concat;
}

// src/web/tags.ts
var HTML_TAGS = [
  "a",
  "abbr",
  "address",
  "area",
  "article",
  "aside",
  "audio",
  "b",
  "base",
  "bdo",
  "bgsound",
  "blink",
  "blockquote",
  "br",
  "button",
  "canvas",
  "caption",
  "cite",
  "code",
  "col",
  "colgroup",
  "datalist",
  "dd",
  "del",
  "details",
  "dfn",
  "div",
  "dl",
  "dt",
  "em",
  "fieldset",
  "figure",
  "footer",
  "form",
  "h1",
  "h2",
  "h3",
  "h4",
  "h5",
  "h6",
  "header",
  "hgroup",
  "hr",
  "i",
  "img",
  "input",
  "ins",
  "kbd",
  "label",
  "legend",
  "li",
  "link",
  "map",
  "mark",
  "marquee",
  "meter",
  "multicol",
  "nav",
  "object",
  "ol",
  "optgroup",
  "option",
  "output",
  "p",
  "pre",
  "progress",
  "q",
  "samp",
  "section",
  "select",
  "small",
  "source",
  "span",
  "strong",
  "style",
  "sub",
  "sup",
  "svg",
  "table",
  "tbody",
  "td",
  "textarea",
  "tfoot",
  "th",
  "thead",
  "time",
  "tr",
  "ul",
  "var",
  "video",
  "wbr"
];

// src/web/jsx.ts
function fixTagName(path) {
  const { name } = path.node.openingElement;
  if (types_default.isJSXIdentifier(name) && !/^[A-Z]/.test(name.name) && !HTML_TAGS.includes(name.name))
    setTagName(path, "div");
}
function getClassName(context, module2) {
  if (!context.props.size && !context.children.size)
    return;
  const { condition, alternate, uid } = context;
  if (typeof condition == "string" || types_default.isStringLiteral(condition))
    return;
  const value = module2 ? types_default.memberExpression(module2, types_default.identifier(uid), false) : types_default.stringLiteral(uid);
  if (!condition)
    return value;
  if (alternate) {
    let alt = getClassName(alternate, module2);
    if (typeof alt === "string")
      alt = types_default.stringLiteral(alt);
    if (alt)
      return types_default.conditionalExpression(condition, value, alt);
  }
  return types_default.logicalExpression("&&", condition, value);
}
function addClassName(path, name, options) {
  const existing = hasProp(path, "className");
  const opening = path.get("openingElement");
  if (typeof name == "string")
    name = types_default.stringLiteral(name);
  if (types_default.isStringLiteral(existing) && types_default.isStringLiteral(name)) {
    existing.value += " " + name.value;
    return;
  }
  if (!existing) {
    opening.pushContainer(
      "attributes",
      types_default.jsxAttribute(
        types_default.jsxIdentifier("className"),
        types_default.isStringLiteral(name) ? name : types_default.jsxExpressionContainer(name)
      )
    );
    return;
  }
  const concat = getHelper("classNames", path, options);
  if (types_default.isCallExpression(existing) && types_default.isIdentifier(existing.callee, { name: concat.name }))
    if (types_default.isStringLiteral(name)) {
      for (const value of existing.arguments)
        if (types_default.isStringLiteral(value)) {
          value.value += " " + name.value;
          return;
        }
    } else {
      existing.arguments.push(name);
      return;
    }
  for (const attr of opening.get("attributes"))
    if (attr.isJSXAttribute() && attr.get("name").isJSXIdentifier({ name: "className" })) {
      attr.node.value = types_default.jsxExpressionContainer(
        types_default.callExpression(concat, [name, existing])
      );
      return;
    }
  throw new Error("Could not insert className");
}

// src/web/macros/index.js
var macros_exports = {};
__export(macros_exports, {
  absolute: () => absolute,
  aspectSize: () => aspectSize,
  background: () => background,
  backgroundImage: () => backgroundImage,
  backgroundSize: () => backgroundSize,
  bg: () => background,
  border: () => border,
  borderB: () => borderBottom,
  borderBottom: () => borderBottom,
  borderL: () => borderLeft,
  borderLeft: () => borderLeft,
  borderR: () => borderRight,
  borderRadius: () => borderRadius,
  borderRight: () => borderRight,
  borderT: () => borderTop,
  borderTop: () => borderTop,
  bottom: () => bottom,
  circle: () => circle,
  family: () => fontFamily,
  fixed: () => fixed,
  flexAlign: () => flexAlign,
  font: () => font,
  fontFamily: () => fontFamily,
  fontSize: () => fontSize,
  gap: () => gap,
  gridArea: () => gridArea,
  gridColumn: () => gridColumn,
  gridColumns: () => gridColumns,
  gridRow: () => gridRow,
  gridRows: () => gridRows,
  height: () => height,
  icon: () => icon,
  image: () => image,
  left: () => left,
  lineHeight: () => lineHeight,
  margin: () => margin,
  marginB: () => marginBottom,
  marginBottom: () => marginBottom,
  marginH: () => marginHorizontal,
  marginHorizontal: () => marginHorizontal,
  marginL: () => marginLeft,
  marginLeft: () => marginLeft,
  marginR: () => marginRight,
  marginRight: () => marginRight,
  marginT: () => marginTop,
  marginTop: () => marginTop,
  marginV: () => marginVertical,
  marginVertical: () => marginVertical,
  maxHeight: () => maxHeight,
  maxSize: () => maxSize,
  maxWidth: () => maxWidth,
  minHeight: () => minHeight,
  minSize: () => minSize,
  minWidth: () => minWidth,
  nToNUnits: () => nToNUnits,
  outlineWidth: () => outlineWidth,
  padding: () => padding,
  paddingH: () => paddingHorizontal,
  paddingHorizontal: () => paddingHorizontal,
  paddingV: () => paddingVertical,
  paddingVertical: () => paddingVertical,
  radius: () => radius,
  relative: () => relative,
  right: () => right,
  shadow: () => shadow,
  size: () => size,
  top: () => top,
  width: () => width
});

// src/web/util.ts
function camelToDash(x) {
  return x.replace(/([A-Z]+)/g, "-$1").toLowerCase();
}
function appendUnitToN(val, unit) {
  if (val === 0)
    return "0";
  if (val === void 0)
    return "";
  if (typeof val == "string" && /\./.test(val) && !isNaN(parseFloat(val)))
    return val + (unit || "em");
  if (typeof val == "number" || parseInt(val) === Number(val))
    return val + (unit || "px");
  return val;
}

// src/web/macros/border.js
function factory(dir) {
  let key = "border";
  if (dir)
    key += dir[0].toUpperCase() + dir.slice(1);
  return (color, width2, style) => {
    if (color == "none" || color == "transparent" || !width2)
      return {
        [key]: color
      };
    return {
      [key]: [
        color || "black",
        style || "solid",
        appendUnitToN(width2 || "1")
      ]
    };
  };
}
var border = factory();
var borderTop = factory("top");
var borderLeft = factory("left");
var borderRight = factory("right");
var borderBottom = factory("bottom");

// src/web/macros/color.js
function background(a) {
  if (Array.isArray(a)) {
    const [head, ...tail] = a;
    switch (head) {
      case "rgb":
      case "rgba": {
        const { value } = rgba(...tail);
        return {
          backgroundColor: value
        };
      }
      case "hsl":
      case "hsla": {
        const { value } = hsla(...tail);
        return {
          backgroundColor: value
        };
      }
    }
  }
  return {
    background: Array.from(arguments)
  };
}
function rgba(r, g, b, a = 1) {
  for (const x of [r, g, b])
    if (typeof x != "number")
      throw new Error("malformed arguments in rgb statement");
  const rgb = [r, g, b].join(",");
  return {
    value: a == 1 ? `rgb(${rgb})` : `rgba(${rgb},${a})`
  };
}
function hsla(h, s, l, a = 1) {
  for (const x of [h, s, l])
    if (typeof x != "number")
      throw new Error("malformed arguments in hsl statement");
  const hsl = [h, s + "%", l + "%"].join(",");
  return {
    value: a == 1 ? `hsl(${hsl})` : `hsla(${hsl},${a})`
  };
}

// src/web/macros/flex.js
var DIRECTIONS = {
  right: "row",
  left: "row-reverse",
  up: "column-reverse",
  down: "column",
  row: null,
  column: null,
  "row-reverse": null,
  "column-reverse": null
};
function flexAlign() {
  const style = {
    display: "flex"
  };
  for (const arg of arguments) {
    if (arg in DIRECTIONS) {
      style.flexDirection = DIRECTIONS[arg] || arg;
    } else if (arg == "center") {
      style.justifyContent = "center";
      style.alignItems = "center";
    } else
      style.justifyContent = arg;
  }
  return style;
}

// src/web/macros/grid.js
function gridArea(...args) {
  if (args.length == 2)
    return {
      gridRow: recombineSlash(args[0]),
      gridColumn: recombineSlash(args[1])
    };
}
function gridRow(...args) {
  return {
    gridRow: recombineSlash(args)
  };
}
function gridColumn(...args) {
  return {
    gridColumn: recombineSlash(args)
  };
}
function gridRows(...args) {
  return {
    display: "grid",
    gridTemplateRows: recombineTemplate(args)
  };
}
function gridColumns(...args) {
  return {
    display: "grid",
    gridTemplateColumns: recombineTemplate(args)
  };
}
function recombineSlash(args) {
  args = [].concat(args);
  if (args[0] !== "-")
    return args[0];
  let layer = args;
  let x = "";
  while (true) {
    x = x + " / " + layer[2];
    if (Array.isArray(layer[1]))
      layer = layer[1];
    else
      return layer[1] + x;
  }
}
function recombineTemplate(args) {
  return args.map(formatGridValue).join(" ");
}
function formatGridValue(x) {
  if (typeof x == "number")
    return x + "px";
  if (typeof x !== "string")
    throw new Error("Unexpected value for grid");
  if (/^\d+\.\d+$/.test(x))
    return x + "fr";
  if (x == "min" || x == "max")
    return x + "-content";
  return x;
}

// src/web/macros/image.js
function image(a) {
  return {
    backgroundImage: `url("${a}")`
  };
}
function requireExpression(value) {
  return {
    type: "CallExpression",
    callee: { type: "Identifier", name: "require" },
    arguments: [
      { type: "StringLiteral", value }
    ]
  };
}
function backgroundImage(from) {
  if (/^\.\.?\//.test(from))
    return {
      backgroundImage: {
        type: "TemplateLiteral",
        expressions: [
          requireExpression(from)
        ],
        quasis: [
          {
            type: "TemplateElement",
            value: {
              raw: "url(",
              cooked: "url("
            },
            tail: false
          },
          {
            type: "TemplateElement",
            value: {
              raw: ")",
              cooked: ")"
            },
            tail: true
          }
        ]
      }
    };
  if (typeof from == "object" && !from.named)
    return {
      backgroundImage: from
    };
  return {
    backgroundImage: Array.from(arguments)
  };
}
function icon(mask, color) {
  if (!mask)
    return;
  if (!mask.includes(".svg"))
    mask = mask.concat(".svg");
  const output = {
    WebkitMaskImage: `url("${mask}")`
  };
  if (color)
    output.bg = color;
  return output;
}

// src/web/macros/margins.js
function margin(a1) {
  let margin2;
  if (arguments.length == 1 && a1 == "auto" || a1 == "none" || / /.test(a1))
    margin2 = a1;
  else
    margin2 = Array.from(arguments).map((x) => appendUnitToN(x)).join(" ");
  return {
    margin: margin2
  };
}
function padding(a1) {
  let padding2;
  if (arguments.length == 1 && a1 == "auto" || a1 == "none" || / /.test(a1))
    padding2 = a1;
  else
    padding2 = Array.from(arguments).map((x) => appendUnitToN(x)).join(" ");
  return {
    padding: padding2
  };
}
function marginTop(...args) {
  return {
    marginTop: appendUnitToN(...args)
  };
}
function marginLeft(...args) {
  return {
    marginLeft: appendUnitToN(...args)
  };
}
function marginRight(...args) {
  return {
    marginRight: appendUnitToN(...args)
  };
}
function marginBottom(...args) {
  return {
    marginBottom: appendUnitToN(...args)
  };
}
function paddingHorizontal(a, b) {
  return {
    paddingLeft: a,
    paddingRight: b || a
  };
}
function paddingVertical(a, b) {
  return {
    paddingTop: a,
    paddingBottom: b || a
  };
}
function marginHorizontal(a, b) {
  return {
    marginLeft: a,
    marginRight: b || a
  };
}
function marginVertical(a, b) {
  return {
    marginTop: a,
    marginBottom: b || a
  };
}

// src/web/macros/position.js
function absolute(...args) {
  return position("absolute", ...args);
}
function fixed(...args) {
  return position("fixed", ...args);
}
function relative() {
  return {
    position: "relative"
  };
}
var INVERSE = {
  top: "bottom",
  left: "right",
  right: "left",
  bottom: "top"
};
function position(kind, a, b = 0, c = b, d) {
  const out = {
    position: kind,
    top: b,
    left: c,
    right: c,
    bottom: b
  };
  if (a == "fill")
    return out;
  if (typeof a == "string") {
    const [k1, k2] = a.split("-");
    if (k2) {
      if (k1 == "fill")
        delete out[INVERSE[k2]];
      else
        for (const dir of [k1, k2])
          delete out[INVERSE[dir]];
      return out;
    }
  }
  let top2;
  let left2;
  let right2;
  let bottom2;
  switch (arguments.length - 1) {
    case 0:
      a = 0;
    case 1:
      top2 = right2 = bottom2 = left2 = a;
      break;
    case 2:
      top2 = bottom2 = a;
      left2 = right2 = b;
      break;
    case 3:
      top2 = a;
      bottom2 = c;
      left2 = right2 = b;
      break;
    case 4:
      top2 = a;
      right2 = b;
      bottom2 = c;
      left2 = d;
      break;
    default:
      throw new Error("Too many arguments for css 4-way value.");
  }
  return {
    position: kind,
    top: top2,
    right: right2,
    bottom: bottom2,
    left: left2
  };
}

// src/web/macros/radius.js
var CORNER_MATRIX = {
  top: [1, 1, 0, 0],
  left: [1, 0, 0, 1],
  right: [0, 1, 1, 0],
  bottom: [0, 0, 1, 1]
};
function radius(dir, r1, r2) {
  let value = "";
  if (dir == "round")
    value = 999;
  else if (r1 === void 0)
    value = dir;
  else if (typeof dir == "string") {
    let [d1, d2] = dir.split("-");
    let matrix = CORNER_MATRIX[d1];
    if (d2)
      matrix = matrix.map((dir2, i) => {
        return CORNER_MATRIX[d2][i] ? dir2 : 0;
      });
    const radii = matrix.map((b) => {
      return (b ? r1 : r2) || 0;
    });
    value = radii.map(appendUnitToN).join(" ");
  }
  return {
    borderRadius: appendUnitToN(value)
  };
}
function circle(a) {
  return {
    borderRadius: a / 2,
    size: a
  };
}

// src/web/macros/scalar.js
function nToNUnits(value, unit) {
  if (value == "fill")
    value = "100%";
  if (value.named) {
    unit = value.named;
    value = value.inner[0];
  }
  return [
    appendUnitToN(value, unit)
  ];
}
var gap = nToNUnits;
var top = nToNUnits;
var left = nToNUnits;
var right = nToNUnits;
var bottom = nToNUnits;
var width = nToNUnits;
var height = nToNUnits;
var maxWidth = nToNUnits;
var maxHeight = nToNUnits;
var minWidth = nToNUnits;
var minHeight = nToNUnits;
var fontSize = nToNUnits;
var lineHeight = nToNUnits;
var outlineWidth = nToNUnits;
var borderRadius = nToNUnits;
var backgroundSize = nToNUnits;

// src/web/macros/shadow.js
function shadow(color, radius2 = 10, x = 2, y = x) {
  const boxShadow = color == "intial" || color == "none" ? color : `${x}px ${y}px ${radius2}px ${color}`;
  return {
    boxShadow
  };
}

// src/web/macros/size.js
function withPrefix(prefix) {
  const width2 = prefix ? `${prefix}Width` : "width";
  const height2 = prefix ? `${prefix}Height` : "height";
  return function size2(x, y, unit) {
    if (typeof y == "string" && typeof x == "number") {
      unit = y;
      y = null;
    }
    return {
      [width2]: [x, unit],
      [height2]: [y || x, unit]
    };
  };
}
var size = withPrefix();
var minSize = withPrefix("min");
var maxSize = withPrefix("max");
function aspectSize(x, y, unit) {
  const y2 = Math.abs(y);
  if (y2 && y2 < 1)
    if (y <= 0)
      y = x * y2;
    else {
      y = x;
      x *= y2;
    }
  return {
    width: [x, unit],
    height: [y || x, unit]
  };
}

// src/web/macros/font.js
function font(...args) {
  const output = {};
  for (const x of args)
    if (x % 100 === 0)
      output.fontWeight = x;
    else if (isNaN(x) === false)
      output.fontSize = x;
    else if (typeof x === "string")
      output.fontFamily = x;
  return output;
}
function fontFamily(...args) {
  return {
    fontFamily: Array.from(args).map(quoteIfWhitespace).join(", ")
  };
}
function quoteIfWhitespace(font2) {
  return ~font2.indexOf(" ") ? `"${font2}"` : font2;
}

// src/web/styles.ts
function isInUse(context) {
  return context.props.size > 0;
}
function byPriority(a, b) {
  return depth(a) - depth(b);
}
function toCss(context) {
  const css = [];
  for (let [name, value] of context.props)
    css.push("  " + toCssProperty(name, value));
  const select = toSelector(context);
  const style = css.join("\n");
  return `${select} {
${style}
}`;
}
function toCssProperty(name, value) {
  const property2 = name.replace(/^\$/, "--").replace(/([A-Z]+)/g, "-$1").toLowerCase();
  if (Array.isArray(value))
    value = value.map((value2) => {
      if (typeof value2 == "string" && /^\$/.test(value2))
        return `var(--${camelToDash(value2.slice(1))})`;
      return value2;
    });
  return `${property2}: ${value.join(" ")};`;
}
function toSelector(context) {
  let { parent, condition, uid } = context;
  if (typeof condition === "string")
    return toSelector(context.parent) + condition;
  let selector = "";
  while (parent) {
    if (parent instanceof Context && parent.condition) {
      selector = toSelector(parent) + " ";
      break;
    }
    parent = parent.parent;
  }
  return selector += "." + uid;
}
function depth(context) {
  let depth2 = 0;
  do {
    if (context.path.isFunction())
      break;
    else
      depth2 += /^[A-Z]/.test(context.uid) ? 2 : 1;
  } while (context = context.parent);
  return depth2;
}

// src/web/index.ts
function Preset(_compiler, options = {}) {
  const { cssModule } = options;
  let cssObject;
  Object.assign(types_default, _compiler.types);
  return {
    plugins: [
      [plugin_default, {
        ...options,
        macros: [
          macros_exports,
          ...options.macros || []
        ],
        apply(path, using, state) {
          const used = new Set(using);
          let forward;
          for (const define of used) {
            const className = getClassName(define, cssObject);
            if (className)
              addClassName(path, className, options);
            if (define.path.isFunction())
              forward = define.path;
          }
          for (const context of used) {
            const { styles } = state.file.metadata;
            const key = toSelector(context);
            context.children.forEach((x) => used.add(x));
            styles.set(key, context);
          }
          if (forward) {
            spreadProps(path, componentProps(forward));
            if (hasProp(path, "className"))
              addClassName(path, componentProp(path, "className"), options);
          }
          fixTagName(path);
        }
      }],
      [{
        visitor: {
          Program: {
            enter(path, state) {
              if (cssModule)
                cssObject = uniqueIdentifier(path.scope, "css");
              Object.defineProperties(state.file.metadata, {
                styles: {
                  enumerable: true,
                  value: /* @__PURE__ */ new Map()
                },
                css: {
                  enumerable: true,
                  get() {
                    return Array.from(this.styles.values()).filter(isInUse).sort(byPriority).map(toCss).join("\n");
                  }
                }
              });
            },
            exit(path, state) {
              const { styles } = state.file.metadata;
              if (cssObject && cssModule && styles.size)
                path.unshiftContainer("body", types_default.importDeclaration(
                  [types_default.importDefaultSpecifier(cssObject)],
                  types_default.stringLiteral(cssModule)
                ));
            }
          }
        }
      }]
    ]
  };
}
var web_default = Preset;
